@@ -1,113 +1,20 @@
 import datetime
-import errno
-import socket
 import threading
 import time
 import warnings
-from itertools import chain, imap
-from redis.exceptions import ConnectionError, ResponseError, InvalidResponse, WatchError
-from redis.exceptions import RedisError, AuthenticationError
+from itertools import chain, imap, islice, izip
+from redis.connection import ConnectionPool, Connection
+from redis.exceptions import (
+    AuthenticationError,
+    ConnectionError,
+    DataError,
+    PubSubError,
+    RedisError,
+    ResponseError,
+    WatchError,
+)
 
 
-class ConnectionPool(threading.local):
-    "Manages a list of connections on the local thread"
-    def __init__(self):
-        self.connections = {}
-
-    def make_connection_key(self, host, port, db):
-        "Create a unique key for the specified host, port and db"
-        return '%s:%s:%s' % (host, port, db)
-
-    def get_connection(self, host, port, db, password, socket_timeout):
-        "Return a specific connection for the specified host, port and db"
-        key = self.make_connection_key(host, port, db)
-        if key not in self.connections:
-            self.connections[key] = Connection(
-                host, port, db, password, socket_timeout)
-        return self.connections[key]
-
-    def get_all_connections(self):
-        "Return a list of all connection objects the manager knows about"
-        return self.connections.values()
-
-
-class Connection(object):
-    "Manages TCP communication to and from a Redis server"
-    def __init__(self, host='localhost', port=6379, db=0, password=None,
-                 socket_timeout=None):
-        self.host = host
-        self.port = port
-        self.db = db
-        self.password = password
-        self.socket_timeout = socket_timeout
-        self._sock = None
-        self._fp = None
-
-    def connect(self, redis_instance):
-        "Connects to the Redis server if not already connected"
-        if self._sock:
-            return
-        try:
-            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            sock.settimeout(self.socket_timeout)
-            sock.connect((self.host, self.port))
-        except socket.error, e:
-            # args for socket.error can either be (errno, "message")
-            # or just "message"
-            if len(e.args) == 1:
-                error_message = "Error connecting to %s:%s. %s." % \
-                    (self.host, self.port, e.args[0])
-            else:
-                error_message = "Error %s connecting %s:%s. %s." % \
-                    (e.args[0], self.host, self.port, e.args[1])
-            raise ConnectionError(error_message)
-        sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
-        self._sock = sock
-        self._fp = sock.makefile('r')
-        redis_instance._setup_connection()
-
-    def disconnect(self):
-        "Disconnects from the Redis server"
-        if self._sock is None:
-            return
-        try:
-            self._sock.close()
-        except socket.error:
-            pass
-        self._sock = None
-        self._fp = None
-
-    def send(self, command, redis_instance):
-        "Send ``command`` to the Redis server. Return the result."
-        self.connect(redis_instance)
-        try:
-            self._sock.sendall(command)
-        except socket.error, e:
-            if e.args[0] == errno.EPIPE:
-                self.disconnect()
-            if isinstance(e.args, basestring):
-                _errno, errmsg = 'UNKNOWN', e.args
-            else:
-                _errno, errmsg = e.args
-            raise ConnectionError("Error %s while writing to socket. %s." % \
-                (_errno, errmsg))
-
-    def read(self, length=None):
-        """
-        Read a line from the socket is length is None,
-        otherwise read ``length`` bytes
-        """
-        try:
-            if length is not None:
-                return self._fp.read(length)
-            return self._fp.readline()
-        except socket.error, e:
-            self.disconnect()
-            if e.args and e.args[0] == errno.EAGAIN:
-                raise ConnectionError("Error while reading from socket: %s" % \
-                    e.args[1])
-        return ''
-
 def list_or_args(command, keys, args):
     # returns a single list combining keys and args
     # if keys is not a list or args has items, issue a
@@ -141,7 +48,7 @@ def timestamp_to_datetime(response):
     return datetime.datetime.fromtimestamp(response)
 
 def string_keys_to_dict(key_string, callback):
-    return dict([(key, callback) for key in key_string.split()])
+    return dict.fromkeys(key_string.split(), callback)
 
 def dict_merge(*dicts):
     merged = {}
@@ -156,23 +63,28 @@ def parse_info(response):
             return value
         sub_dict = {}
         for item in value.split(','):
-            k, v = item.split('=')
+            k, v = item.rsplit('=', 1)
             try:
                 sub_dict[k] = int(v)
             except ValueError:
                 sub_dict[k] = v
         return sub_dict
     for line in response.splitlines():
-        key, value = line.split(':')
-        try:
-            info[key] = int(value)
-        except ValueError:
-            info[key] = get_value(value)
+        if line and not line.startswith('#'):
+            key, value = line.split(':')
+            try:
+                if '.' in value:
+                    info[key] = float(value)
+                else:
+                    info[key] = int(value)
+            except ValueError:
+                info[key] = get_value(value)
     return info
 
 def pairs_to_dict(response):
     "Create a dict given a list of key/value pairs"
-    return dict(zip(response[::2], response[1::2]))
+    it = iter(response)
+    return dict(izip(it, it))
 
 def zset_score_pairs(response, **options):
     """
@@ -181,7 +93,8 @@ def zset_score_pairs(response, **options):
     """
     if not response or not options['withscores']:
         return response
-    return zip(response[::2], map(float, response[1::2]))
+    it = iter(response)
+    return zip(it, imap(float, it))
 
 def int_or_none(response):
     if response is None:
@@ -224,7 +137,7 @@ class Redis(threading.local):
         string_keys_to_dict(
             # these return OK, or int if redis-server is >=1.3.4
             'LPUSH RPUSH',
-            lambda r: isinstance(r, int) and r or r == 'OK'
+            lambda r: isinstance(r, long) and r or r == 'OK'
             ),
         string_keys_to_dict('ZSCORE ZINCRBY', float_or_none),
         string_keys_to_dict(
@@ -236,7 +149,9 @@ class Redis(threading.local):
         string_keys_to_dict('SDIFF SINTER SMEMBERS SUNION',
             lambda r: r and set(r) or set()
             ),
-        string_keys_to_dict('ZRANGE ZRANGEBYSCORE ZREVRANGE', zset_score_pairs),
+        string_keys_to_dict('ZRANGE ZRANGEBYSCORE ZREVRANGE ZREVRANGEBYSCORE',
+            zset_score_pairs
+            ),
         string_keys_to_dict('ZRANK ZREVRANK', int_or_none),
         {
             'BGREWRITEAOF': lambda r: \
@@ -315,8 +230,8 @@ class Redis(threading.local):
     def _execute_command(self, command_name, command, **options):
         subscription_command = command_name in self.SUBSCRIPTION_COMMANDS
         if self.subscribed and not subscription_command:
-            raise RedisError("Cannot issue commands other than SUBSCRIBE and "
-                "UNSUBSCRIBE while channels are open")
+            raise PubSubError("Cannot issue commands other than SUBSCRIBE and "
+                              "UNSUBSCRIBE while channels are open")
         try:
             self.connection.send(command, self)
             if subscription_command:
@@ -339,64 +254,9 @@ class Redis(threading.local):
             **options
             )
 
-    def _parse_response(self, command_name, catch_errors):
-        conn = self.connection
-        response = conn.read()[:-2] # strip last two characters (\r\n)
-        if not response:
-            self.connection.disconnect()
-            raise ConnectionError("Socket closed on remote end")
-
-        # server returned a null value
-        if response in ('$-1', '*-1'):
-            return None
-        byte, response = response[0], response[1:]
-
-        # server returned an error
-        if byte == '-':
-            if response.startswith('ERR '):
-                response = response[4:]
-            raise ResponseError(response)
-        # single value
-        elif byte == '+':
-            return response
-        # int value
-        elif byte == ':':
-            return int(response)
-        # bulk response
-        elif byte == '$':
-            length = int(response)
-            if length == -1:
-                return None
-            response = length and conn.read(length) or ''
-            conn.read(2) # read the \r\n delimiter
-            return response
-        # multi-bulk response
-        elif byte == '*':
-            length = int(response)
-            if length == -1:
-                return None
-            if not catch_errors:
-                return [self._parse_response(command_name, catch_errors)
-                    for i in range(length)]
-            else:
-                # for pipelines, we need to read everything,
-                # including response errors. otherwise we'd
-                # completely mess up the receive buffer
-                data = []
-                for i in range(length):
-                    try:
-                        data.append(
-                            self._parse_response(command_name, catch_errors)
-                            )
-                    except Exception, e:
-                        data.append(e)
-                return data
-
-        raise InvalidResponse("Unknown response type for: %s" % command_name)
-
     def parse_response(self, command_name, catch_errors=False, **options):
         "Parses a response from the Redis server"
-        response = self._parse_response(command_name, catch_errors)
+        response = self.connection.read_response(command_name, catch_errors)
         if command_name in self.RESPONSE_CALLBACKS:
             return self.RESPONSE_CALLBACKS[command_name](response, **options)
         return response
@@ -427,6 +287,7 @@ class Redis(threading.local):
         connection object calls this method to authenticate and select
         the appropriate database.
         """
+        self.subscribed = False
         if self.connection.password:
             if not self.execute_command('AUTH', self.connection.password):
                 raise AuthenticationError("Invalid Password")
@@ -460,7 +321,7 @@ class Redis(threading.local):
     def shutdown(self):
         "Shutdown the server"
         if self.subscribed:
-            raise RedisError("Can't call 'shutdown' from a pipeline'")
+            raise PubSubError("Can't call 'shutdown' when 'subscribed'")
         try:
             self.execute_command('SHUTDOWN')
         except ConnectionError:
@@ -746,21 +607,21 @@ class Redis(threading.local):
         "Returns the type of key ``name``"
         return self.execute_command('TYPE', name)
 
-    def watch(self, name):
+    def watch(self, *names):
         """
-        Watches the value at key ``name``, or None of the key doesn't exist
+        Watches the values at keys ``names``, or None if the key doesn't exist
         """
         if self.subscribed:
-            raise RedisError("Can't call 'watch' from a pipeline'")
+            raise PubSubError("Can't call 'watch' when 'subscribed'")
 
-        return self.execute_command('WATCH', name)
+        return self.execute_command('WATCH', *names)
 
     def unwatch(self):
         """
         Unwatches the value at key ``name``, or None of the key doesn't exist
         """
         if self.subscribed:
-            raise RedisError("Can't call 'unwatch' from a pipeline'")
+            raise PubSubError("Can't call 'unwatch' when 'subscribed'")
 
         return self.execute_command('UNWATCH')
 
@@ -826,17 +687,17 @@ class Redis(threading.local):
         end of the list
         """
         return self.execute_command('LINDEX', name, index)
-    
+
     def linsert(self, name, where, refvalue, value):
         """
         Insert ``value`` in list ``name`` either immediately before or after
         [``where``] ``refvalue``
-        
+
         Returns the new length of the list on success or -1 if ``refvalue``
         is not in the list.
         """
         return self.execute_command('LINSERT', name, where, refvalue, value)
-    
+
     def llen(self, name):
         "Return the length of the list ``name``"
         return self.execute_command('LLEN', name)
@@ -848,7 +709,7 @@ class Redis(threading.local):
     def lpush(self, name, value):
         "Push ``value`` onto the head of the list ``name``"
         return self.execute_command('LPUSH', name, value)
-    
+
     def lpushx(self, name, value):
         "Push ``value`` onto the head of the list ``name`` if ``name`` exists"
         return self.execute_command('LPUSHX', name, value)
@@ -1107,7 +968,7 @@ class Redis(threading.local):
         ``desc`` indicates to sort in descending order.
 
         ``withscores`` indicates to return the scores along with the values.
-            The return type is a list of (value, score) pairs
+        The return type is a list of (value, score) pairs
         """
         if desc:
             return self.zrevrange(name, start, end, withscores)
@@ -1122,10 +983,11 @@ class Redis(threading.local):
         Return a range of values from the sorted set ``name`` with scores
         between ``min`` and ``max``.
 
-        If ``start`` and ``num`` are specified, then return a slice of the range.
+        If ``start`` and ``num`` are specified, then return a slice
+        of the range.
 
         ``withscores`` indicates to return the scores along with the values.
-            The return type is a list of (value, score) pairs
+        The return type is a list of (value, score) pairs
         """
         if (start is not None and num is None) or \
                 (num is not None and start is None):
@@ -1172,13 +1034,35 @@ class Redis(threading.local):
         ``start`` and ``num`` can be negative, indicating the end of the range.
 
         ``withscores`` indicates to return the scores along with the values
-            as a dictionary of value => score
+        as a dictionary of value => score
         """
         pieces = ['ZREVRANGE', name, start, num]
         if withscores:
             pieces.append('withscores')
         return self.execute_command(*pieces, **{'withscores': withscores})
 
+    def zrevrangebyscore(self, name, max, min,
+            start=None, num=None, withscores=False):
+        """
+        Return a range of values from the sorted set ``name`` with scores
+        between ``min`` and ``max`` in descending order.
+
+        If ``start`` and ``num`` are specified, then return a slice
+        of the range.
+
+        ``withscores`` indicates to return the scores along with the values.
+        The return type is a list of (value, score) pairs
+        """
+        if (start is not None and num is None) or \
+                (num is not None and start is None):
+            raise RedisError("``start`` and ``num`` must both be specified")
+        pieces = ['ZREVRANGEBYSCORE', name, max, min]
+        if start is not None and num is not None:
+            pieces.extend(['LIMIT', start, num])
+        if withscores:
+            pieces.append('withscores')
+        return self.execute_command(*pieces, **{'withscores': withscores})
+
     def zrevrank(self, name, value):
         """
         Returns a 0-based value indicating the descending rank of
@@ -1207,9 +1091,7 @@ class Redis(threading.local):
     def _zaggregate(self, command, dest, keys, aggregate=None):
         pieces = [command, dest, len(keys)]
         if isinstance(keys, dict):
-            items = keys.items()
-            keys = [i[0] for i in items]
-            weights = [i[1] for i in items]
+            keys, weights = keys.keys(), keys.values()
         else:
             weights = None
         pieces.extend(keys)
@@ -1269,6 +1151,8 @@ class Redis(threading.local):
         Sets each key in the ``mapping`` dict to its corresponding value
         in the hash ``name``
         """
+        if not mapping:
+            raise DataError("'hmset' with 'mapping' of length 0")
         items = []
         for pair in mapping.iteritems():
             items.extend(pair)
@@ -1427,7 +1311,7 @@ class Pipeline(Redis):
                 "pipeline execution")
         # Run any callbacks for the commands run in the pipeline
         data = []
-        for r, cmd in zip(response, commands):
+        for r, cmd in izip(response, commands):
             if not isinstance(r, Exception):
                 if cmd[0] in self.RESPONSE_CALLBACKS:
                     r = self.RESPONSE_CALLBACKS[cmd[0]](r, **cmd[2])
@@ -1462,6 +1346,9 @@ class Pipeline(Redis):
     def select(self, *args, **kwargs):
         raise RedisError("Cannot select a different database from a pipeline")
 
+class LockError(RedisError):
+    "Errors thrown from the Lock"
+    pass
 
 class Lock(object):
     """
@@ -1472,7 +1359,7 @@ class Lock(object):
     multiple clients play nicely together.
     """
 
-    LOCK_FOREVER = 2**31+1 # 1 past max unix time
+    LOCK_FOREVER = float(2**31+1) # 1 past max unix time
 
     def __init__(self, redis, name, timeout=None, sleep=0.1):
         """
@@ -1495,6 +1382,8 @@ class Lock(object):
         self.acquired_until = None
         self.timeout = timeout
         self.sleep = sleep
+        if self.timeout and self.sleep > self.timeout:
+            raise LockError("'sleep' must be less than 'timeout'")
 
     def __enter__(self):
         return self.acquire()
@@ -1518,15 +1407,16 @@ class Lock(object):
                 timeout_at = unixtime + timeout
             else:
                 timeout_at = Lock.LOCK_FOREVER
+            timeout_at = float(timeout_at)
             if self.redis.setnx(self.name, timeout_at):
                 self.acquired_until = timeout_at
                 return True
             # We want blocking, but didn't acquire the lock
             # check to see if the current lock is expired
-            existing = long(self.redis.get(self.name) or 1)
+            existing = float(self.redis.get(self.name) or 1)
             if existing < unixtime:
                 # the previous lock is expired, attempt to overwrite it
-                existing = long(self.redis.getset(self.name, timeout_at) or 1)
+                existing = float(self.redis.getset(self.name, timeout_at) or 1)
                 if existing < unixtime:
                     # we successfully acquired the lock
                     self.acquired_until = timeout_at
@@ -1539,7 +1429,7 @@ class Lock(object):
         "Releases the already acquired lock"
         if self.acquired_until is None:
             raise ValueError("Cannot release an unlocked lock")
-        existing = long(self.redis.get(self.name) or 1)
+        existing = float(self.redis.get(self.name) or 1)
         # if the lock time is in the future, delete the lock
         if existing >= self.acquired_until:
             self.redis.delete(self.name)
